import inspect
import unittest

import numpy as np

from pydtnn.backends.cpu.utils.im2row_nhwc_cython import im2row_nhwc_cython
from pydtnn.libs.libconvwinograd import ConvWinograd
from pydtnn.tests.abstract.common import verbose_test, D
from pydtnn.tests.abstract.conv_common import ConvCommonTestCase
from pydtnn.utils import print_with_header
from pydtnn.utils.tensor import TensorFormat

# if (kh, kw) == (2, 2) and (vstride, hstride) == (1, 1) and (vdilation, hdilation) == (1, 1):
# if (kh, kw) == (3, 3) and (vstride, hstride) == (1, 1) and (vdilation, hdilation) == (1, 1):
# if (kh, kw) == (3, 3) and (vstride, hstride) == (1, 1) and (vdilation, hdilation) == (1, 1):
# if (kh, kw) == (5, 5) and (vstride, hstride) == (1, 1) and (vdilation, hdilation) == (1, 1):


class ConvWinogradTestCase(ConvCommonTestCase):
    """
    Tests that conv_winograd leads to the same results as i2c and mm.
    """
    # NOTE: Delete parent test to prevent re-export and re-testing
    global ConvCommonTestCase
    del ConvCommonTestCase

    @classmethod
    def _compute_both(cls, weights: np.ndarray, x: np.ndarray, biases: np.ndarray | None = None,
                      vpadding=0, hpadding=0, vstride=1, hstride=1,
                      vdilation=1, hdilation=1) -> tuple[np.ndarray, np.ndarray]:
        c, kh, kw, kn = weights.shape
        # b, c, h, w = x.shape
        cw_biases = biases.copy() if biases is not None else None
        conv_winograd_result: np.ndarray = cls._compute(weights, x, biases=cw_biases,
                                                        kh=kh, kw=kw,
                                                        vpadding=vpadding, hpadding=hpadding,
                                                        vstride=vstride, hstride=hstride,
                                                        vdilation=vdilation, hdilation=hdilation)
        conv_winograd_result: np.ndarray = conv_winograd_result.reshape((-1, kn), copy=False)

        n, h, w, c = x.shape

        ho = (h + 2 * vpadding - vdilation * (kh - 1) - 1) // vstride + 1
        wo = (w + 2 * hpadding - hdilation * (kw - 1) - 1) // hstride + 1

        dim_n = n * ho * wo
        dim_c = c * kh * kw

        x_c: np.ndarray = np.zeros(shape=(dim_n, dim_c), dtype=x.dtype)

        im2row_nhwc_cython(x, x_c,
                           kh, kw, ho, wo,
                           vpadding, hpadding,
                           vstride, hstride,
                           vdilation, hdilation)
        w_c = weights.reshape((-1, kn), copy=False)
        im2row_mm_result: np.ndarray = np.matmul(x_c, w_c)
        if biases is not None:
            np.add(im2row_mm_result, biases.reshape((-1, kn), copy=False), out=im2row_mm_result, dtype=im2row_mm_result.dtype)
        if verbose_test():
            print_with_header("{} conv_winograd_result".format(inspect.stack()[1][3]), conv_winograd_result)
            print("Shape: ", conv_winograd_result.shape,
                  " Sum: ", conv_winograd_result.sum(),
                  " Min: ", conv_winograd_result.min(),
                  " Max: ", conv_winograd_result.max())
            print_with_header("{} im2row_mm_result".format(inspect.stack()[1][3]), im2row_mm_result)
            print("Shape: ", im2row_mm_result.shape,
                  " Sum: ", im2row_mm_result.sum(),
                  " Min: ", im2row_mm_result.min(),
                  " Max: ", im2row_mm_result.max())
            print("---")
            print("Maximum difference: ",
                  max([abs(x - y) for x, y in zip(conv_winograd_result.flatten(), im2row_mm_result.flatten())]))
            print("---")
        return conv_winograd_result, im2row_mm_result

    @staticmethod
    def _compute(weights: np.ndarray, x: np.ndarray, biases: np.ndarray | None = None, kh=1, kw=1, vpadding=0, hpadding=0, vstride=1, hstride=1, vdilation=1, hdilation=1):
        return ConvWinograd(kh, kw, vstride, hstride, vdilation, hdilation, debug=verbose_test(), tensor_format=TensorFormat.NHWC).conv_winograd_nhwc(weights, x, biases, vpadding, hpadding, vstride, hstride, vdilation, hdilation)

    @staticmethod
    def _get_config() -> D:
        return D(h=100, w=100, kh=3, kw=3)

    @unittest.skip("Winograd only supports stride 1x1")
    def test_raise_on_different_strides(self):
        raise NotImplementedError()

    @unittest.skip("Winograd only supports stride 1x1")
    def test_with_different_stride(self):
        raise NotImplementedError()

    @unittest.skip("Winograd only supports stride 1x1")
    def test_with_different_strides(self):
        raise NotImplementedError()

    @unittest.skip("Winograd only supports dilation 1x1")
    def test_with_different_dilation(self):
        raise NotImplementedError()

    @unittest.skip("Winograd only supports a subset of kernel sizes")
    def test_alexnet_layers(self):
        raise NotImplementedError()



"""
PyDTNN Layer base class
"""


from pydtnn.layer_base import LayerBase
from pydtnn.utils import find_component
from pydtnn.utils.constants import Array


class LayerError(ValueError):
    pass


class ParameterException(LayerError):
    pass


class Layer[T: Array](LayerBase[T]):
    pass


def select(name: str) -> type[Layer]:
    assert __package__, "Package not found!"
    return find_component(__package__, name)



from pydtnn.backends.cpu.layers.abstract.conv_2d_standard import Conv2DStandardCPU
from pydtnn.layers.conv_2d_relu import Conv2DRelu
from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.utils.constants import ArrayShape

import numpy as np

# Next no inspection is because Conv2D _backward_depthwise and _backward_pointwise being considered as abstract methods
# noinspection PyAbstractClass


class Conv2DReluCPU(Conv2DRelu[np.ndarray], Conv2DStandardCPU):

    # NOTE: The "__init__" method is being made (more or less) in Model (in _apply_layer_fusion) and in FusedLayerMixIn.

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None) -> None:
        super().initialize(prev_shape, x)
        self.forward = {"_forward_cg_nchw": self._forward_nchw_cg,
                        "_forward_cg_nhwc": self._forward_nhwc_cg,
                        "_forward_cw_nchw": self._forward_nchw_cw}[self.forward.__name__]
        self.backward = self._backward

    def _forward_nchw_cg(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses the convGemm + Relu"""

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONVGEMM)
        res: np.ndarray = self.cg.conv_gemm_nchw(self.weights, x, out=None,
                                              vpadding=self.vpadding, hpadding=self.hpadding,
                                              vstride=self.vstride, hstride=self.hstride,
                                              vdilation=self.vdilation, hdilation=self.hdilation,
                                              biases=self.biases, relu=True)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(res, dtype=self.model.dtype, order='C', copy=None)

    def _forward_nhwc_cg(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses the convGemm + Relu"""

        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONVGEMM)
        res: np.ndarray = self.cg.conv_gemm_nhwc(self.weights, x, out=None,
                                              vpadding=self.vpadding, hpadding=self.hpadding,
                                              vstride=self.vstride, hstride=self.hstride,
                                              vdilation=self.vdilation, hdilation=self.hdilation,
                                              biases=self.biases, relu=True)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return np.asarray(res, dtype=self.model.dtype, order='C', copy=None)

    def _forward_nchw_cw(self, x: np.ndarray) -> np.ndarray:
        """Version of the forward function that uses the convWinograd + Relu"""
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CONVGEMM)
        y: np.ndarray = self.cw.conv_winograd_nchw(self.weights, x, self.biases,
                                                vpadding=self.vpadding, hpadding=self.hpadding,
                                                vstride=self.vstride, hstride=self.hstride,
                                                vdilation=self.vdilation, hdilation=self.hdilation,
                                                relu=True)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)

        return np.asarray(y, dtype=self.model.dtype, order='C', copy=None)

    def _backward(self, dy: np.ndarray) -> np.ndarray:
        raise NotImplementedError("Use a real backwards variant!")



try:
    import pycuda.driver as drv  #type: ignore
except Exception as e:
    pass

from typing import TYPE_CHECKING

from pydtnn.tracers.simple_tracer import SimpleTracer

if TYPE_CHECKING:
    from pympi.MPI import Comm as MPI_COMM
else:
    from types import ModuleType
    MPI_COMM = ModuleType


class SimpleTracerGPU(SimpleTracer):
    """
    SimpleTracerGPU
    """

    def __init__(self, tracing: bool, output_filename: str, comm: MPI_COMM | None):
        super().__init__(tracing, output_filename, comm)
        self.event_vars = []
        # Attributes that will be initialized later
        self.stream: drv.Stream = None

    def _get_start_end_event(self) -> tuple:
        if len(self.event_vars) == 0:
            self.event_vars.append((drv.Event(), drv.Event()))
        return self.event_vars.pop()

    def _release_start_end_event(self, start, end):
        self.event_vars.append((start, end))

    def _emit_event(self, evt_type_val: int, evt_val: int, stream=None):
        """This method will be called only if tracing is enabled"""
        if stream is None:
            stream = self.stream
        if evt_val != 0:
            start, end = self._get_start_end_event()
            self.pending_events.append((evt_type_val, evt_val, start, end))
            start.record(stream=stream)
        else:
            if len(self.pending_events) == 0:
                raise RuntimeError("Received an 'End' event but there are no pending events!")
            if self.pending_events[-1][0] != evt_type_val:
                raise RuntimeError("Received an 'End' event for a different event type than expected!")
            _evt_type_val, _evt_val, start, end = self.pending_events.pop()
            end.record(stream=stream)
            end.synchronize()
            evt_time = start.time_till(end) * 1e-3
            self._release_start_end_event(start, end)
            previous_calls, previous_time = self.events[_evt_type_val][_evt_val]
            self.events[_evt_type_val][_evt_val] = [previous_calls + 1, previous_time + evt_time]

    def _emit_nevent(self, evt_type_val_list: list, evt_val_list: list, stream=None):
        """This method will be called only if tracing is enabled"""
        zipped_list = list(zip(evt_type_val_list, evt_val_list))
        if evt_val_list[0] == 0:
            zipped_list = reversed(zipped_list)
        for evt_type_val, evt_val in zipped_list:
            self.emit_event(evt_type_val, evt_val, stream)

    def set_stream(self, stream):
        self.stream = stream



"""Library test group"""
from warnings import warn

# ConvGemm
try:
    from pydtnn.tests.conv_gemm import ConvGemmTestCase
    from pydtnn.tests.conv_2d_conv_gemm import Conv2DConvGemmTestCase
    from pydtnn.tests.model_conv_gemm import ModelConvGemmTestCase
except Exception:
    warn("ConvGemm not available, skiping tests!")

# ConvWinograd
try:
    from pydtnn.tests.conv_winograd import ConvWinogradTestCase
except Exception:
    warn("ConvWinograd not available, skiping tests!")

# ConvDirect
try:
    from pydtnn.tests.conv_direct import ConvDirectTestCase
except Exception:
    warn("ConvDirect not available, skiping tests!")



import resource
from collections import defaultdict

import numpy as np


class PerformanceCounter:
    TRAINING, TESTING = range(2)

    def __init__(self):
        self._times_record = defaultdict(lambda: defaultdict(lambda: []))
        self._batch_sizes_record = defaultdict(lambda: defaultdict(lambda: []))
        self._memory_record = defaultdict(lambda: defaultdict(lambda: []))

    # -------------------------------
    #  Public methods and properties
    # -------------------------------

    def add_training_time_and_batch_size(self, epoch, elapsed_time, batch_size):
        self._add_time_and_batch_size(self.TRAINING, epoch, elapsed_time, batch_size)

    def add_testing_time_and_batch_size(self, test_round, elapsed_time, batch_size):
        self._add_time_and_batch_size(self.TESTING, test_round, elapsed_time, batch_size)

    @property
    def training_throughput(self):
        return self._throughput(self.TRAINING)

    @property
    def training_throughput_only_last_half_of_each_epoch(self):
        return self._throughput(self.TRAINING, last_half=True)

    @property
    def num_epochs(self):
        return len(self._batch_sizes_record[self.TRAINING].keys())

    @property
    def num_evaluations(self):
        return len(self._batch_sizes_record[self.TESTING].keys())

    @property
    def training_time(self):
        return self._time(self.TRAINING)

    @property
    def training_time_estimated_from_last_half_of_each_epoch(self):
        return self._time(self.TRAINING, last_half=True)

    @property
    def training_maximum_memory(self):
        return self._maximum_memory(self.TRAINING)

    @property
    def training_mean_memory(self):
        return self._mean_memory(self.TRAINING)

    @property
    def testing_throughput(self):
        return self._throughput(self.TESTING)

    @property
    def testing_time(self):
        return self._time(self.TESTING)

    @property
    def testing_maximum_memory(self):
        return self._maximum_memory(self.TESTING)

    @property
    def testing_mean_memory(self):
        return self._mean_memory(self.TESTING)

    def print_report(self):
        if self.num_epochs > 0:
            print(" -------------------------------------")
            print("| Performance counter training report |")
            print(" -------------------------------------")
            print(f'Training time (from model): {self.training_time:5.4f} s')
            print(f'Training time per epoch (from model): '
                  f'{self.training_time / self.num_epochs:5.4f} s')
            print(f'Training throughput (from model): {self.training_throughput:5.4f} samples/s')
            print(f'Training time (from model, estimated from last half of each epoch): '
                  f'{self.training_time_estimated_from_last_half_of_each_epoch:5.4f} s')
            print(f'Training throughput (from model, from last half of each epoch): '
                  f'{self.training_throughput_only_last_half_of_each_epoch:5.4f} samples/s')
            print(f'Training maximum memory allocated: '
                  f'{self.training_maximum_memory / 1024:.2f} MiB')
            print(f'Training mean memory allocated: '
                  f'{self.training_mean_memory / 1024:.2f} MiB')

        if self.num_evaluations > 0:
            print(" ------------------------------------")
            print("| Performance counter testing report |")
            print(" ------------------------------------")
            print(f'Testing time (from model): {self.testing_time / self.num_evaluations:5.4f} s')
            print(f'Testing throughput (from model): {self.testing_throughput:5.4f} samples/s')
            print(f'Testing maximum memory allocated: ',
                  f'{self.testing_maximum_memory / 1024:.2f} MiB')
            print(f'Testing mean memory allocated: ',
                  f'{self.testing_mean_memory / 1024:.2f} MiB')

    # -------------------------------
    #  Private methods
    # -------------------------------

    def _add_time_and_batch_size(self, where, epoch, elapsed_time, batch_size):
        self._times_record[where][epoch].append(elapsed_time)
        self._batch_sizes_record[where][epoch].append(batch_size)
        mem = (resource.getrusage(resource.RUSAGE_SELF)[2]
               + resource.getrusage(resource.RUSAGE_CHILDREN)[2])
        self._memory_record[where][epoch].append(mem)  # KiB in GNU/Linux

    def _time(self, where, last_half=False):
        return self._sum(self._times_record[where].values(), last_half)

    @staticmethod
    def _sum(arrays, last_half):
        # When last_half is True, the total size is estimated from the last half steps of each epoch size
        if not last_half:
            records_per_epoch = [np.sum(array) for array in arrays]
        else:
            records_per_epoch = []
            for array in arrays:
                array_last_half = array[len(array) // 2:]
                if len(array_last_half) > 0:
                    records_per_epoch.append(np.sum(array_last_half) * len(array) / len(array_last_half))
        return np.sum(records_per_epoch)

    def _size(self, where, last_half=False):
        return self._sum(self._batch_sizes_record[where].values(), last_half)

    def _throughput(self, where, last_half=False):
        return self._size(where, last_half) / self._time(where, last_half)

    def _maximum_memory(self, where):
        if where == self.TRAINING:
            maximum_memory_per_epoch = [np.max(m_array) for m_array in self._memory_record[where].values()]
            return np.max(maximum_memory_per_epoch)
        else:
            # Consider only the first evaluation
            maximum_memory_first_evaluation = np.max(self._memory_record[where][0])
            return maximum_memory_first_evaluation

    def _mean_memory(self, where):
        if where == self.TRAINING:
            mean_memory_per_epoch = [np.mean(m_array)
                                     for m_array in self._memory_record[where].values()]
            return np.mean(mean_memory_per_epoch)
        else:
            # Consider only the first evaluation
            mean_memory_first_evaluation = np.mean(self._memory_record[where][0])
            return mean_memory_first_evaluation



import unittest

from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.conv_2d_batch_normalization import Conv2DBatchNormalization
from pydtnn.model import Model
from pydtnn.backends.cpu.layers.conv_2d import Conv2DCPU
from pydtnn.tests.abstract.common import D
from pydtnn.tests.abstract.common import Params
from pydtnn.tests.abstract.conv_2d_common import Conv2DCommonTestCase
from pydtnn.utils.tensor import TensorFormat
from pydtnn.utils.initializers import glorot_uniform, zeros


class Conv2DBatchNormalizationTestCase(Conv2DCommonTestCase):
    """
    Tests that Conv2D+BatchNormalization leads to the same results than Conv2DBatchNormalization
    """
    # NOTE: Delete parent test to prevent re-export and re-testing
    global Conv2DCommonTestCase
    del Conv2DCommonTestCase

    @staticmethod
    def _get_layers(d: D, deconv=False, trans=False) -> tuple[Conv2DCPU, Conv2DCPU]:
        params = Params()
        params.tensor_format = TensorFormat.NCHW.upper()
        params.batch_size = d.b
        params.conv_variant = "gemm"
        model = Model(**vars(params))
        model.mode = Model.Mode.TRAIN

        conv2d = Conv2D(nfilters=d.kn, filter_shape=(d.kh, d.kw),
                        padding=(d.vpadding, d.hpadding),
                        stride=(d.vstride, d.hstride),
                        dilation=(d.vdilation, d.hdilation),
                        use_bias=True, weights_initializer=glorot_uniform, biases_initializer=zeros)
        bn = BatchNormalization()
        chain = ConcatenationBlock([
            conv2d,
            bn
        ])
        shape = (d.c, d.h, d.w)
        chain.init_backend_from_model(model)
        chain.initialize(prev_shape=shape, x=None)

        from_parent = bn.__dict__ | conv2d.__dict__
        fuse = Conv2DBatchNormalization(from_parent=from_parent)
        fuse.init_backend_from_model(model)
        fuse.__dict__.update(from_parent)
        fuse.initialize(prev_shape=shape, x=None)

        # Set the same initial weights and biases to both layers
        fuse.weights = conv2d.weights.copy()
        fuse.biases = conv2d.biases.copy()

        return chain, fuse

    @staticmethod
    def _set_state(layer: Conv2D, weights) -> None:
        if isinstance(layer, ConcatenationBlock):
            layer.paths[0][0].weights = weights.copy()
        else:
            layer.weights = weights.copy()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_larger_handmade_array_stride3(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_larger_handmade_array_stride2(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_handmade_array(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_even_larger_handmade_array_stride3_filter1x2(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_even_larger_handmade_array_stride3_filter1x1(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_defaults(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_handmade_array_stride2(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_even_larger_handmade_array_stride3(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_even_larger_handmade_array_stride12(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_alexnet_imagenet_first_conv2d(self):
        raise NotImplementedError()

    @unittest.skip("Backward not implemented")
    def test_forward_backward_alexnet_cifar10_first_conv2d(self):
        raise NotImplementedError()



from collections.abc import Sequence

from pydtnn.activations.relu import Relu
from pydtnn.activations.softmax import Softmax
from pydtnn.layers.average_pool_2d import AveragePool2D
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layer_base import LayerBase
from pydtnn.utils.initializers import he_uniform


def densenet161(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=input_shape))

    blocks, growth_rate = [6, 12, 36, 24], 48  # DenseNet161

    reduction = 0.5
    num_planes = 2 * growth_rate

    _(Conv2D(nfilters=num_planes, filter_shape=(3, 3), padding=1, use_bias=False, weights_initializer=he_uniform))

    for i, nblocks in enumerate(blocks):
        for j in range(nblocks):
            _(ConcatenationBlock(
                [
                    BatchNormalization(),
                    Relu(),
                    Conv2D(nfilters=4 * growth_rate, filter_shape=(1, 1), use_bias=False,
                           weights_initializer=he_uniform),
                    BatchNormalization(),
                    Relu(),
                    Conv2D(nfilters=growth_rate, filter_shape=(3, 3), padding=1, use_bias=False,
                           weights_initializer=he_uniform)
                ], []))

        num_planes += nblocks * growth_rate

        if i < len(blocks) - 1:
            num_planes = int(num_planes * reduction)
            _(BatchNormalization())
            _(Relu())
            _(Conv2D(nfilters=num_planes, filter_shape=(1, 1), use_bias=False, weights_initializer=he_uniform))
            _(AveragePool2D(pool_shape=(2, 2), stride=2))

    _(BatchNormalization())
    _(Relu())
    _(AveragePool2D(pool_shape=(4, 4)))
    _(Flatten())
    _(FC(shape=output_shape, activation=Softmax))

    return model



from collections.abc import Sequence

from pydtnn.activations.relu import Relu
from pydtnn.activations.softmax import Softmax
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.dropout import Dropout
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layer_base import LayerBase

from pydtnn.layer_base import LayerBase
from pydtnn.utils.initializers import he_uniform
from pydtnn.layers.max_pool_2d import MaxPool2D


def vgg11bn(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=input_shape))
    conv_pattern = [[1, 64], [1, 128], [2, 256], [2, 512], [2, 512]]
    for nlayers, nfilters in conv_pattern:
        for layer in range(nlayers):
            _(Conv2D(nfilters=nfilters, filter_shape=(3, 3), padding=1, stride=1, weights_initializer=he_uniform))
            _(BatchNormalization())
            _(Relu())
        _(MaxPool2D(pool_shape=(2, 2), stride=2))
    _(Flatten())
    _(Dropout(rate=0.5))
    _(FC(shape=(512,), weights_initializer=he_uniform))
    _(BatchNormalization())
    _(Relu())
    _(Dropout(rate=0.5))
    _(FC(shape=(512,), weights_initializer=he_uniform))
    _(BatchNormalization())
    _(Relu())
    _(FC(shape=output_shape, activation=Softmax, weights_initializer=he_uniform))

    return model



from typing import TYPE_CHECKING

import numpy as np

from pydtnn.datasets.dataset import Dataset
from pydtnn.utils import random


if TYPE_CHECKING:
    from pydtnn.model import Model

TRAIN_NSAMPLES = 10000
TEST_NSAMPLES = 2000
INPUT_SHAPE = (1,)
OUTPUT_SHAPE = (1,)


class IWSLT(Dataset):
    """
    IWSLT Dataset

    NOTE: Source unclear
    TODO: Load original dataset

    The IWSLT 2017 Multilingual Task addresses text translation,
    including zero-shot translation, with a single MT system across
    all directions including English, German, Dutch, Italian and
    Romanian. As unofficial task, conventional bilingual text
    translation is offered between English and Arabic, French,
    Japanese, Chinese, German and Korean.

    Source (SHA1): https://huggingface.co/datasets/IWSLT/iwslt2017
    ac7581060e37d197d7fd2f09bb3bf1f8a2d57ddb iwslt_corto.txt
    a3166ce94db1e8e1cff3757f927c7f821229cfe6 iwslt.txt
    8f3d50b3ca9bae54e5cda80d84e5cdb5e33a4b4a wiki_split.txt

    Normalize (z-score):
    offset: ???
    scale:  ???
    """

    def __init__(self, model: "Model", embedl=512, max_sentence=512, split_token="<translation>", force_test_as_validation=False, debug=False):
        super().__init__(model, TRAIN_NSAMPLES, TEST_NSAMPLES, INPUT_SHAPE, OUTPUT_SHAPE, force_test_as_validation=force_test_as_validation, debug=debug)

        self.split_token = split_token
        self.max_sentence = max_sentence
        self.embedl = embedl
        self.train_path = self.model.dataset_path
        self.test_path = self.model.dataset_path
        self.test_as_validation = force_test_as_validation
        self.dtype = self.model.dtype
        self.model = model
        self.lang1 = self.model.dataset_lang
        self.lang2 = self.model.dataset_lang2

    def _init_actual_data(self):
        # Actual
        self.load_data(self.train_path, self.lang1, self.lang2)
        self.make_train_val_partitions()

        # Synthetic
        # self.train_val_nsamples = 10000
        # self.train_nsamples = None
        # self.lines1 = self.lines2 = self.lines1_train = self.lines2_train = self.lines1_val = self.lines2_val = self.lines1_test = self.lines2_test = None
        # self.make_train_val_partitions()
        # # self.src_embeddings = random.random(*self.train_val_nsamples, 1, self.max_sentence, self.embedl)).astype(dtype=self.dtype)
        # # self.tgt_embeddings = random.random(*self.train_val_nsamples, 1, self.max_sentence, self.embedl)).astype(dtype=self.dtype)
        # # self.src_mask = np.zeros((self.train_val_nsamples,1,self.max_sentence),dtype=bool)
        # # self.tgt_mask = np.zeros((self.train_val_nsamples,1,self.max_sentence),dtype=bool)
        # self.src_embeddings = random.random((1000, 1, self.max_sentence, self.embedl)).astype(dtype=self.dtype)
        # self.tgt_embeddings = random.random((1000, 1, self.max_sentence, self.embedl)).astype(dtype=self.dtype)
        # # self.src_mask = np.zeros((1000,1,self.max_sentence),dtype=bool)
        # # self.tgt_mask = np.zeros((1000,1,self.max_sentence),dtype=bool)

    def load_data(self, file, lang1, lang2):
        self.dictionary1 = self.get_dictionary(lang1)
        self.dictionary2 = self.get_dictionary(lang2)
        file = open(self.train_path, "r")
        lines = file.readlines()
        file.close()
        lines = [line.replace("\n", "") for line in lines]
        self.train_val_nsamples = len(lines)
        self.train_nsamples = None
        self.lines1 = [line.split(self.split_token)[0] for line in lines]
        self.lines2 = [line.split(self.split_token)[1] for line in lines]
        # sos = None # <sos>
        # eos = None # <eos>
        # pad = ' '
        # if sos is not None and eos is not None: # No encuentro los eos y sos en spacy
        #     sos = dictionary1(sos).vector
        #     eos = dictionary1(eos).vector
        # pad = dictionary(pad).vector  # No hace falta si inicializamos a 0s

    def get_dictionary(self, language):
        import spacy
        table = {
            "en": "en_core_web_md",
            "de": "de_core_news_md"
        }
        if language in table:
            language = table[language]
        return spacy.load(language)

    def make_train_val_partitions(self):
        val_split = self.model.validation_split
        if self.train_nsamples == None:
            s = np.arange(self.train_val_nsamples)
            if self.model.augment_shuffle:
                random.shuffle(s)
            self.train_nsamples = int(self.train_val_nsamples * (1-val_split) // 1)
            self.train_indices = s[:self.train_nsamples]
            self.val_indices = s[self.train_nsamples:]
            self.val_nsamples = len(self.val_indices)
            self.test_nsamples = self.val_nsamples
            # Make test partition
            if self.test_as_validation:
                self.test_indices = self.val_indices
            # print(len(self.train_indices), len(self.val_indices))
            if self.lines1 is not None:
                self.lines1_train = [self.lines1[i] for i in self.train_indices]
                self.lines2_train = [self.lines2[i] for i in self.train_indices]
                self.lines1_val = [self.lines1[i] for i in self.val_indices]
                self.lines2_val = [self.lines2[i] for i in self.val_indices]
                if self.test_as_validation:
                    self.lines1_test = self.lines1_val
                    self.lines2_test = self.lines2_val
                else:
                    self.lines1_test = [self.lines1[i] for i in self.test_indices]
                    self.lines2_test = [self.lines2[i] for i in self.test_indices]

    def _actual_data_generator(self, part):
        match part:
            case Dataset.Part.TRAIN:
                lines1, lines2 = self.lines1_train, self.lines2_train
            case Dataset.Part.VAL:
                lines1, lines2 = self.lines1_val, self.lines2_val
            case Dataset.Part.TEST:
                lines1, lines2 = self.lines1_test, self.lines2_test

        batch_size = self.model.batch_size

        for i in range(self.train_val_nsamples // batch_size):
            window = (i*batch_size, (i+1)*batch_size)
            src_embeddings = np.zeros((batch_size, self.max_sentence, self.embedl), dtype=np.float32)
            tgt_embeddings = np.zeros((batch_size, self.max_sentence, self.embedl), dtype=np.float32)
            src_mask = np.zeros((batch_size, 1, self.max_sentence), dtype=bool)
            tgt_mask = np.zeros((batch_size, self.max_sentence, self.max_sentence), dtype=bool)
            for i, doc in enumerate(self.dictionary1.pipe(lines1[window[0]:window[1]])):
                for j, word in enumerate(doc):
                    src_embeddings[i, j] = word.vector
                    src_mask[i, 0, j] = 1
            for i, doc in enumerate(self.dictionary2.pipe(lines2[window[0]:window[1]])):
                for j, word in enumerate(doc):
                    tgt_embeddings[i, j] = word.vector
                    tgt_mask[i, j, 0:j+1] = [1] * (j+1)

            x = [src_embeddings, src_mask, tgt_embeddings, tgt_mask]
            y = tgt_embeddings
            yield x, y

    def _synthetic_data_generator(self):
        batch_size = self.model.batch_size
        rank = self.model.rank

        for i in range(self.train_nsamples//batch_size):
            # window = (i * batch_size + rank * batch_size, i * batch_size + (rank + 1) * batch_size)
            window = (0 * batch_size + rank * batch_size, 0 * batch_size + (rank + 1) * batch_size)
            # x = [self.src_embeddings[window[0]:window[1]], self.src_mask[window[0]:window[1]], self.tgt_embeddings[window[0]:window[1]], self.tgt_mask[window[0]:window[1]]]
            # y = self.tgt_embeddings[window[0]:window[1]]
            x = [self.src_embeddings[window[0]:window[1]], self.tgt_embeddings[window[0]:window[1]]]
            y = self.tgt_embeddings[window[0]:window[1]]
            yield x, y

    # === Preprocess ===
    def preprocess(self, original_file_lang1, lang1, original_file_lang2, lang2, destination_file="IWSLT.txt"):
        self.load_data(destination_file, lang1, lang2)
        file = open(original_file_lang1, "r")
        lines1 = file.readlines()
        file.close()
        file = open(original_file_lang2, "r")
        lines2 = file.readlines()
        file.close()
        if len(lines1) != len(lines2):
            print("Los archivos tienen un numero de muestras distintas, {} y {}".format(len(lines1), len(lines2)))
            return -1

        file = open(self.file, "w")
        for i in range(len(lines1)):
            line1 = lines1[i].replace("\n", "")
            line2 = lines2[i].replace("\n", "")
            if len(self.dictionary1(line1)) <= self.max_sentence and len(self.dictionary2(line2)) <= self.max_sentence:
                file.write(line1 + self.split_token + line2 + "\n")
        file.close()

    def xml_to_txt(self, file_in, file_out):
        file = open(file_in, "r")
        # lines = file.read().splitlines()
        lines = file.readlines()
        file.close()
        nlines_in = len(lines)
        lines = [line.replace("\n", "") for line in lines if "<" not in line]
        file = open(file_out, "w")
        file.writelines("\n".join(lines))
        file.close()
        nlines_out = len(lines)
        print("{} lines in, {} lines out".format(nlines_in, nlines_out))
    # === Preprocess ===



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def max_pool_2d_fwd_nchw_cython[T: _npDT](x: _npDT_4Dims[T],
                                          y: _npDT_4Dims[T],
                                          idx_max: _np.ndarray[tuple[int, int, int, int], _np.int32],
                                          kh: int, kw: int, ho: int, wo: int,
                                          vpadding: int, hpadding: int,
                                          vstride: int, hstride: int,
                                          vdilation: int, hdilation: int,
                                          minval: _npDT) -> None:
    """
    Args:
        x (npDT_4Dims): 4-dimensinal array where the input data is stored.
        y (npDT_4Dims): 4-dimensinal array where the output data will be stored.
        idx_max (np.ndarray[tuple[int, int, int, int], np.int32]): 4-dimensinal array where the index of the maximum values will be stored.
        kh (int): Kernel's heigth.
        kw (int): Kernel's width.
        ho (int): Output's heigth.
        wo (int): Output's width.
        vpadding (int): Vertical padding value.
        hpadding (int): Horizontal padding value.
        vstride (int): Vertical stride value.
        hstride (int): Horizontal stride value.
        vdilation (int): Vertical dilation value.
        hdilation (int): Horizontal dilation value.
        minval (npDT): minum value the selected type can have.
    Returns:
        Nothing. The output is stored in "y" and "idx_max"
    """


def max_pool_2d_bwd_nchw_cython[T: _npDT](dy: _npDT_4Dims[T],
                                          idx_max: _np.ndarray[tuple[int, int, int, int], _np.int32],
                                          dx: _npDT_4Dims[T],
                                          n: int, h: int, w: int, c: int,
                                          kh: int, kw: int, ho: int, wo: int,
                                          vpadding: int, hpadding: int,
                                          vstride: int, hstride: int,
                                          vdilation: int, hdilation: int) -> None:
    """
    Args:
        dy (npDT_4Dims): 4-dimensinal array where the input data will be stored.
        idx_max (np.ndarray[tuple[int, int, int, int], np.int32]): 4-dimensinal array where the index of the maximum values will be stored.
        dx (npDT_4Dims): 4 dimensional ndarray where the gradient is stored.
        n (int): Number of samples.
        h (int): Sample's heigth.
        w (int): Sample's width.
        c (int): Sample's channels.
        kh (int): Kernel's heigth.
        kw (int): Kernel's width.
        ho (int): Output's heigth.
        wo (int): Output's width.
        vpadding (int): Vertical padding value.
        hpadding (int): Horizontal padding value.
        vstride (int): Vertical stride value.
        hstride (int): Horizontal stride value.
        vdilation (int): Vertical dilation value.
        hdilation (int): Horizontal dilation value.
    Returns:
        Nothing. The output is stored in "dx".
    """



from pydtnn.layers.layer import Layer
from pydtnn.utils.constants import Array


class AbstractBlockLayer[T: Array](Layer[T]):

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.paths = []
        for path in args:
            self.paths.append(path)
        self.is_block_layer = True
        self.out_shapes: list[tuple[int, ...]] = []

    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)
        self.initialize_block_layer()

    def initialize_block_layer(self):
        for p_i, p in enumerate(self.paths):
            prev_shape = self.prev_shape
            x = self.x
            for i, layer in enumerate(p):
                layer.init_backend_from_model(self.model)
                layer.parent_layer = self
                layer.initialize(prev_shape, x)
                x = layer.y
                if p_i == 0 and (len(p) - 1) == i:
                    self.y = x
                prev_shape = layer.shape
                self.fwd_time += layer.fwd_time
                self.bwd_time += layer.bwd_time
                self.nparams += layer.nparams
            self.out_shapes.append(prev_shape)
        self.shape = self.out_shapes[0]

    def update_weights(self, optimizer):
        for p in self.paths:
            for layer in p:
                layer.update_weights(optimizer)

    def reduce_weights_async(self, gradient=True):
        for p in self.paths:
            for layer in p:
                layer.reduce_weights_async(gradient=gradient)

    def wait_allreduce_async(self, gradient=True):
        for p in self.paths:
            for layer in p:
                layer.wait_allreduce_async(gradient=gradient)

    def reduce_weights_sync(self, gradient=True):
        for p in self.paths:
            for layer in p:
                layer.reduce_weights_sync(gradient=gradient)

    def print_in_convdirect_format(self):
        for p in self.paths:
            for layer in p:
                layer.print_in_convdirect_format()



from collections.abc import Sequence

from pydtnn.activations.softmax import Softmax
from pydtnn.activations.relu import Relu
from pydtnn.layers.average_pool_2d import AveragePool2D
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layer_base import LayerBase
from pydtnn.utils.initializers import he_uniform
from pydtnn.layers.max_pool_2d import MaxPool2D


def googlenet(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=input_shape))

    _(Conv2D(nfilters=192, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform))
    _(BatchNormalization())
    _(Relu())

    inception_blocks = [[64, 96, 128, 16, 32, 32],
                        [128, 128, 192, 32, 96, 64],
                        [],
                        [192, 96, 208, 16, 48, 64],
                        [160, 112, 224, 24, 64, 64],
                        [128, 128, 256, 24, 64, 64],
                        [112, 144, 288, 32, 64, 64],
                        [256, 160, 320, 32, 128, 128],
                        [],
                        [256, 160, 320, 32, 128, 128],
                        [384, 192, 384, 48, 128, 128]]

    for layout in inception_blocks:
        if not layout:
            _(MaxPool2D(pool_shape=(3, 3), stride=2, padding=1))
        else:
            n1x1, n3x3red, n3x3, n5x5red, n5x5, pool_planes = layout
            _(ConcatenationBlock(
                [
                    # 1x1 conv branch
                    Conv2D(nfilters=n1x1, filter_shape=(1, 1), weights_initializer=he_uniform),
                    BatchNormalization(),
                    Relu()
                ],
                [  # 1x1 conv -> 3x3 conv branch
                    Conv2D(nfilters=n3x3red, filter_shape=(1, 1), weights_initializer=he_uniform),
                    BatchNormalization(),
                    Relu(),
                    Conv2D(nfilters=n3x3, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform),
                    BatchNormalization(),
                    Relu()
                ],
                [  # 1x1 conv -> 5x5 conv branch
                    Conv2D(nfilters=n5x5red, filter_shape=(1, 1), weights_initializer=he_uniform),
                    BatchNormalization(),
                    Relu(),
                    Conv2D(nfilters=n5x5, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform),
                    BatchNormalization(),
                    Relu(),
                    Conv2D(nfilters=n5x5, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform),
                    BatchNormalization(),
                    Relu()
                ],
                [  # 3x3 pool -> 1x1 conv branch
                    MaxPool2D(pool_shape=(3, 3), stride=1, padding=1),
                    Conv2D(nfilters=pool_planes, filter_shape=(1, 1), weights_initializer=he_uniform),
                    BatchNormalization(),
                    Relu()
                ]))

    _(AveragePool2D(pool_shape=(8, 8), stride=1))  # Global average pooling 2D
    _(Flatten())
    _(FC(shape=(1024,)))
    _(BatchNormalization())
    _(Relu())
    _(FC(shape=output_shape, activation=Softmax))

    return model



import numpy as np

from pydtnn.backends.gpu.metrics.metric import MetricGPU
from pydtnn.metrics.precision import Precision
from pydtnn.utils.constants import DTYPE2CTYPE
from pycuda.compiler import SourceModule  # type: ignore
from pycuda.driver import Function  # type: ignore

from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU


class PrecisionGPU(Precision[TensorGPU], MetricGPU):

    def initialize(self) -> None:
        super().initialize()
        target_classes = self.model.output_shape[0]
        self.precision = TensorGPU.create_zeros_tensor(shape=(1, ), dtype=np.dtype(np.int32), 
                                                       tensor_format=self.model.tensor_format, cudnn_dtype=self.model.cudnn_dtype)
        self.local_precision = TensorGPU.create_zeros_tensor(shape=(target_classes, ), dtype=np.dtype(np.int32), 
                                                             tensor_format=self.model.tensor_format, cudnn_dtype=self.model.cudnn_dtype)

    def __init_gpu_kernel__(self) -> Function:
        _name = "binary_confusion_matrix"
        code = """
        //#define TRUE_POSITIVE  {{0,0}}
        #define TRUE_POSITIVE_0  0
        #define TRUE_POSITIVE_1  0
        
        //#define FALSE_POSITIVE {{1,0}}
        #define FALSE_POSITIVE_0 1
        #define FALSE_POSITIVE_1 0

        #define SHIFT_POINTER_CM(p, label, i, j, num_i, num_j) p + (((label * num_i + i) * num_j) + j)
        
        __global__ void {name}({T} *precision, int *cm, {T} *local_precision, const int num_classes)
        {{
            int label, idx, true_positive, false_positive, div;

            int base_idx = blockIdx.x * blockDim.x + threadIdx.x;
            const int workers = blockDim.x * gridDim.x;

            for(idx = base_idx; idx < num_classes; idx += workers)
            {{
                *(local_precision + idx) = 0;
                true_positive = (*(SHIFT_POINTER_CM(cm, label, TRUE_POSITIVE_0, TRUE_POSITIVE_1, 2, 2)));
                false_positive = (*(SHIFT_POINTER_CM(cm, label, FALSE_POSITIVE_0, FALSE_POSITIVE_1, 2, 2)));

                div = true_positive + false_positive;

                (*(local_precision + idx)) += ({T}) (div == 0 ? 0 : (true_positive / div));
            }}
            
            // Accumulating the local values into the output's tensor.
            if (base_idx == 0)
            {{
                for(idx = 0; label < num_classes; label++)
                    (*local_precision) += *(local_precision + idx);

                (*precision) = ({T}) ((*local_precision) / num_classes);
            }}
        }}
        """.format(
            T = DTYPE2CTYPE[self.model.dtype],
            name = _name
        )
        module = SourceModule(code).get_function(_name)

        return module
    #---

    def compute(self, y_pred: TensorGPU, y_targ: TensorGPU) -> float:

        target_classes = self.model.output_shape[0]

        target_classes = np.int32(target_classes)

        self.precision.fill(0)
        self.local_precision.fill(0)

        self.kernel(self.precision.ary, self.conf_matrix_metric.conf_matrix.ary, 
                    self.local_precision.ary, target_classes,
                    grid=self.grid, block=self.block,
                    stream=self.model.stream)

        return self.precision.ary.get()[0]



# Typing related (or non important) imports
from typing import Dict, Any, Callable, Tuple, List
from pydtnn.converters.pytorch2pydtnn.layers.activation import LeakyReLU, LogSigmoid, ReLU, ReLU6, Sigmoid, Softmax, Tanh
from pydtnn.converters.pytorch2pydtnn.layers.convolutional import Conv2d
from pydtnn.converters.pytorch2pydtnn.layers.dropout import Dropout
from pydtnn.converters.pytorch2pydtnn.layers.functions import adaptive_avg_pool_2d, add, concat, flatten, relu, sigmoid, softmax, log, tanh
from pydtnn.converters.pytorch2pydtnn.layers.linear import Linear
from pydtnn.converters.pytorch2pydtnn.layers.normalization import BatchNorm2d
from pydtnn.converters.pytorch2pydtnn.layers.pooling import AdaptiveAvgPool2d, AvgPool2d, MaxPool2d
from pydtnn.converters.pytorch2pydtnn.layers.utility import Flatten
from pydtnn.layer_base import LayerBase

# Functionality imports


# ------------------- #
# ---- CONSTANTS ---- #
# ------------------- #
ARGUMENTS = "arguments"
PARAMETERS = "parameters"
LAYERS = "layers"
EQUIVALENT_LAYERS = "equivalent_layers"
OPERATION_VAR = "operation_var"
TRANSPOSE_WEIGHTS_LAYERS = ["Linear"]  # There are layers that put the weigths in the correct order. Theese layers doesn't do it.
REMOVE_WIGHTS_DIMENSIONS = [("Conv2d", (0))]  # Name of the layer, tuple of dimensions/axis to remove.

RELU = "relu"
ADP_AVG_POOL = "adaptive_avg_pool2d"

ADD = "add"
CONCAT = "concat"
CAT = "cat"
FLATTEN = "flatten"
TANH = "tanh"
SOFTMAX = "softmax"
SIGMOID = "sigmoid"
LOG_SIGMOID = "logsigmoid"

ARGS_SEPARATOR = ','
PYTORCH_OUTPUT_SIZE = "output_size"

SPECIAL_CASES = ["torchvision_models_googlenet_GoogLeNetOutputs"]
# SPECIAL CASES:
# -> torchvision_models_googlenet_GoogLeNetOutputs: is a "named tuple". If both aux layers exist and it is not expected their outputs, the actual output is only the FC's one.
# END SPECIAL CASES

# -- END CONSTANTS -- #
# ------------------- #

# ------------------- #
# ---- FUNCTIONS ---- #
# ------------------- #


def not_implemented(name: str) -> Callable:
    # Normal usage of this: switch_pytorch_pydtnn([not_implemented_layer_name])(args)
    def _not_implemented(args: Dict[str, Any]) -> None:
        raise NotImplementedError(f"Layer \"{name}\" not implemented - Args received:\n{args} ")
    return _not_implemented


def prepare_pydtnn_arguments(arguments: Dict[str, Any], torch_dict_keys: List[str], pydtnn_dict_keys: List[str]) -> Dict[str, Any]:
    return {pydtnn_key: arguments[torch_key] for torch_key, pydtnn_key in zip(torch_dict_keys, pydtnn_dict_keys) if torch_key in arguments}


def switch_pytorch_pydtnn(name: str) -> Callable[[Dict[str, Any]], LayerBase]:
    # NOTE: name is the result of torch.nn.[layer]._get_name();
    #   if PyTorch change their layer's names, then it's necessary to change the names here.
    match name:
        case "AdaptiveAvgPool2d": return AdaptiveAvgPool2d
        case "AvgPool2d": return AvgPool2d
        case "BatchNorm2d": return BatchNorm2d
        case "Conv2d": return Conv2d
        case "Dropout": return Dropout
        case "Linear": return Linear
        case "MaxPool2d": return MaxPool2d
        case "ReLU": return ReLU
        case "ReLU6": return ReLU6
        case "LeakyReLU": return LeakyReLU
        case "LogSigmoid": return LogSigmoid
        case "Sigmoid": return Sigmoid
        case "Softmax": return Softmax
        case "Tanh": return Tanh
        case "Flatten": return Flatten

        # Not actual PyTorch layers (are torch functions):
        case "Add": return add  # Possible FIXME: if the constants ADD values are changed, change the case in order to have the same value.
        case "Concat": return concat  # Possible FIXME: if the constants CONCAT values are changed, change the case in order to have the same value.
        # Base case:
        case _: return not_implemented(name)


def switch_operation_symbols(op: str) -> str:
    match op:
        case "+":
            op = ADD
        # Base case:
        case _:
            not_implemented(op)("")
            op = "NOT_IMPLEMENTED"
    return op
# --- switch_operation_symbols --- #


def function_operation_to_pydtnn(name: str) -> Callable[[Dict[str, Any]], Tuple[LayerBase, str]]:

    # NOTE: I found impossible to do a switch (match-case) nor a dictionary due the name may be larger than the "key" (e.g.: name = torch.flatten(input, start_dim=0, end_dim=-1); "key" = "flatten")
    if ADD in name:
        op = add
    elif any(pattern in name for pattern in [CONCAT, CAT]):
        op = concat
    elif FLATTEN in name:
        op = flatten
    elif RELU in name:
        # It is not the layer, but the relu operation itself.
        op = relu
    elif ADP_AVG_POOL in name:
        op = adaptive_avg_pool_2d
    elif LOG_SIGMOID in name:
        op = log
    elif SIGMOID in name:
        # NOTE: is important that SIGMOID is after LOG_SIGMOID
        op = sigmoid
    elif SOFTMAX in name:
        op = softmax
    elif TANH in name:
        op = tanh
    # NOTE: If a new function operation handler is implemented, an "elif" must be place before the followin else in order to call the handler of that operation.
    else:
        op = not_implemented(name)
    return op


def get_lists_operations_and_outputs(dict_layers: Dict[str, Tuple[LayerBase, str]], layer_inputs: List[str]) -> Tuple[List[List[LayerBase]], List[str], str]:
    # NOTE: It is assumed that the model will by a feed-forward network
    dict_branch = {}

    # -- Making the "path" of layers for every input -- #

    for inpt in layer_inputs:
        dict_branch[inpt] = dict()
        input_search = inpt
        while input_search is not None:
            # operations: {[output_name]: [operation]}
            op, inp = dict_layers[input_search]
            dict_branch[inpt][input_search] = op
            input_search = inp
        # end while
    # end for

    # -- Searching the first coincidence -- #

    # NOTE: This is the flow of my thougths regarding the approach:
    #  > Sets are not ordered by insertion ==> keep order with enumerate ==>
    #  > ==> braches have different sizes, then the same node may have different order in different branches ==>
    #  > ==> that's true from bottom to top, but from top to bottom the "intersection layers" -the ones to be searched- (the ones that coincide in all branches) must be in the same position in every branch.
    enumerated_reversed_inputs = enumerate(list(dict_branch[layer_inputs[0]].keys())[::-1])
    coincidences = set(enumerated_reversed_inputs)  # NOTE: It is necessary to have a set with elements in order to make an intersection.
    for i in range(1, len(layer_inputs)):
        coincidences = coincidences.intersection(set(enumerate(list(dict_branch[layer_inputs[i]].keys())[::-1])))
    # "Unenumerating" and sorting the intersection, and getting the first coincidence layer.
    #   ==> NOTE: Due the list was sorting in reverse before, now it is necessary to sort it be reverse again (that's why the "-x[0]").
    coincidences = [elem[1] for elem in sorted(coincidences, key=lambda x: -x[0])]
    new_previous_layer = coincidences[0]  # new_previous_layer = PyDTNN concat input

    # -- Trimming the dict and storing the data to be returned -- #

    lists_operations: List[LayerBase] = list()  # List of lists (one list per branch)
    lists_outputs: List[str] = list()  # List of strings (all branches in one list)
    for inpt in layer_inputs:
        # - Trimming the dict - #
        for coincidence in coincidences:
            del dict_branch[inpt][coincidence]
        # ----

        # - Setting the lists of operations and outputs - #
        # NOTE: dict_branch[].values() is reversed ==> It is necesarry to unreverse the layer
        layers = list(dict_branch[inpt].values())[::-1]
        outputs = list(dict_branch[inpt].keys())
        lists_operations.append(layers)  # NOTE: Remember, this is a list of lists (one per branch)
        lists_outputs.extend(outputs)  # NOTE: Remember, this is a list of strings (all branches in one)
        # ----
    # for inpt in layer_inputs end
    return (lists_operations, lists_outputs, new_previous_layer)


def separate_function_params(params: str) -> List[str]:
    # Example: '[layer1_0_bn3,layer1_0_downsample_1]'
    params = params.replace('[', '').replace(']', '')  # Removing non-useful characters
    params = params.split(',')
    return [param.strip() for param in params]  # Removing spaces

# NOTE: This coversor does *not* work in the cases like the following:
# A, B, C, D, E are layers, D and E are layers like concatenation or addition layers.
# A →→ B → D → E
#   ↘→ C →→↑   ↑
#       ↘→→→→→→↑


def get_equivalent_layer(params: List[str], dict_equivalent_layers: Dict[str, str]) -> List[str]:
    equivalent_layers = dict()
    for param in params:
        layer = param
        while layer in dict_equivalent_layers:
            layer = dict_equivalent_layers[layer]
        equivalent_layers[layer] = None
    return list(equivalent_layers.keys())

# -- END FUNCTIONS -- #
# ------------------- #



from model_convertor import convert_model

from typing import Dict, Tuple, Any, Callable

from torch.nn import Module as PyTorch_Model
import torch.nn as nn
import torch

from pydtnn.model import Model as PyDTNN_Model
from pydtnn.layer_base import LayerBase
from pydtnn.utils.best_of import BestOf

from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
try:
    import pycuda.gpuarray as gpuarray
    from pydtnn.libs import libcudnn as cudnn
except BaseException:
    pass

from copy import deepcopy
from math import prod
import numpy as np
from pydtnn.converters.pytorch2pydtnn.common import TRANSPOSE_WEIGHTS_LAYERS
from pydtnn.utils import random

import pydtnn

# CONSTANTS
N = 100
SHAPE = (3, 20, 20)  # NCHW
CONV_IN_CHANNELS = SHAPE[0]  # Shape format: CHW
CONV_OUT_CHANNELS = 64  # = PyTorch's Number filters
CONV_KERNEL_SIZE = (2, 2)
LINEAR_IN_FEATURES = SHAPE[2]
LINEAR_OUT_FEATURES = prod((SHAPE[0], SHAPE[1]))
BATCH_NORM_IN_FEATURES = SHAPE[0]
POOL_SIZE = (2, 2)
ADAPTIVE_AVG_POOL_OUT_SHAPE = (6, 6)

PYTORCH_LAYER_WEIGHTS = "weight"
PYTORCH_LAYER_BIASES = "bias"

# setting random seed
SEED = 1234
torch.manual_seed(SEED)
random.seed(SEED)
# ---

THRESHOLD = 1e-4
DTYPE = np.float32

KWARGS = {
    "model_name": None,
    # "dataset": None,
    # "dataset_name": None,
    "evaluate_only": True,
    "parallel": "data",
    "tensor_format": "NCHW",  # "NCHW" # "NHWC",
    "loss_func": "categorical_cross_entropy",
    "enable_gpu": False,  # True,
    "omm": None,
    "dtype": DTYPE,
    "tracing": False,
    "tracer_output": "",
    "batch_size": N
}

TYPES_DATA_CUDA = {np.float64: "CUDNN_DATA_DOUBLE",
                   np.float32: "CUDNN_DATA_FLOAT",
                   np.int8: "CUDNN_DATA_INT8",
                   np.int32: "CUDNN_DATA_INT32"}

DICT_SUPPORTED_LAYERS: Dict[str, Tuple[nn.Module, float]] = {
    # Activations:
    "LogSigmoid": (nn.LogSigmoid(), 1e-5),  # PyTorch is more precise ==> it can differ in elements below "e-08"
    "ReLU": (nn.ReLU(), 1e-5),
    "ReLU6": (nn.ReLU6(), 1e-5),
    "LeakyReLU": (nn.LeakyReLU(negative_slope=6), 1e-5),
    "Sigmoid": (nn.Sigmoid(), 1e-5),
    "Softmax": (nn.Softmax(), 1e-5),
    "Tanh": (nn.Tanh(), 1e-5),
    # Convolutional layers:
    "Conv2d": (nn.Conv2d(CONV_IN_CHANNELS, CONV_OUT_CHANNELS, CONV_KERNEL_SIZE), 2e-3),  # PyTorch is more precise ==> it can differ in elements below "e-03"
    # Dropout layers:
    "Dropout": (nn.Dropout(), 1e10),  # It varies due the chosen distribution. In p=0, p=1 and testing mode they have the same results.
    # Linear layers:
    # "Linear": (nn.Linear(LINEAR_IN_FEATURES, LINEAR_OUT_FEATURES), 2e-3),
    # Normalization layers:
    "BatchNorm2d": (nn.BatchNorm2d(BATCH_NORM_IN_FEATURES), 1e-5),
    "Flatten": (nn.Flatten(), 1e-5),
    # Pooling layers:
    "MaxPool2d": (nn.MaxPool2d(POOL_SIZE), 1e-5),
    "AvgPool2d": (nn.AvgPool2d(POOL_SIZE), 1e-5),
    "AdaptiveAvgPool2d": (nn.AdaptiveAvgPool2d(ADAPTIVE_AVG_POOL_OUT_SHAPE), 1e-5),
}
# END CONSTANTS


def print_model_reports(model: PyDTNN_Model):
    # Print performance counter report
    model.perf_counter.print_report()
    # Print BestOf report
    if model.enable_best_of:
        print()
        BestOf.print_report()


class TEST_PyTorch_Model(PyTorch_Model):

    def __init__(self, layer):
        super().__init__()
        self.layer = layer
        print(f"self.layer: {self.layer}")

    def forward(self, x):
        return self.layer(x)


class Addition_Test_PyTorch_Model(PyTorch_Model):

    def __init__(self):
        super().__init__()
        self.op0: nn.Module = DICT_SUPPORTED_LAYERS["AdaptiveAvgPool2d"][0]
        self.op1: nn.Module = DICT_SUPPORTED_LAYERS["MaxPool2d"][0]
        self.op2: nn.Module = DICT_SUPPORTED_LAYERS["AvgPool2d"][0]
        self.act: nn.Module = DICT_SUPPORTED_LAYERS["Tanh"][0]

    def forward(self, x):
        dict_forwards = dict()
        ro = self.op0(x)
        dict_forwards["AvgPool2d"] = ro
        r1 = self.op1(ro)
        dict_forwards["MaxPool2d"] = r1
        r2 = self.op2(ro)
        dict_forwards["AvgPool2d"] = r2
        res = r1 + r2
        dict_forwards["ADD"] = res
        res = self.act(res)
        dict_forwards["Tanh"] = res
        return (res, dict_forwards)


class Concat_Test_PyTorch_Model(PyTorch_Model):

    def __init__(self):
        super().__init__()
        self.op0: nn.Module = DICT_SUPPORTED_LAYERS["AdaptiveAvgPool2d"][0]
        self.op1: nn.Module = DICT_SUPPORTED_LAYERS["MaxPool2d"][0]
        self.op2: nn.Module = DICT_SUPPORTED_LAYERS["AvgPool2d"][0]
        self.activation1: nn.Module = DICT_SUPPORTED_LAYERS["Sigmoid"][0]
        self.activation2: nn.Module = DICT_SUPPORTED_LAYERS["Softmax"][0]
        self.act: nn.Module = DICT_SUPPORTED_LAYERS["Tanh"][0]

    def forward(self, x):
        dict_forwards = dict()
        ro = self.op0(x)
        dict_forwards["AdaptiveAvgPool2d"] = ro
        r1 = self.op1(ro)
        dict_forwards["MaxPool2d"] = r1
        r1 = self.activation1(r1)
        dict_forwards["Sigmoid"] = r1
        r2 = self.op2(ro)
        dict_forwards["AvgPool2d"] = r2
        r2 = self.activation2(r2)
        dict_forwards["Softmax"] = r2
        res = torch.concat([r1, r2], dim=1)
        dict_forwards["CONCAT"] = res
        res = self.act(res)
        dict_forwards["Tanh"] = res
        return (res, dict_forwards)


def are_all_zeros(diff: np.ndarray) -> bool:
    return not diff.any()


def are_all_below_threshold(diff: np.ndarray, threshold: float = THRESHOLD) -> bool:
    return np.all(diff < threshold)


def forward_pydtnn_model(model: PyDTNN_Model, dataset: np.ndarray | TensorGPU) -> np.ndarray | TensorGPU:
    y: np.ndarray | TensorGPU = dataset

    for i in range(1, len(model.layers)):  # NOTE - Remember: Layer 0 is the Input layer and it's ignored
        layer: LayerBase = model.layers[i]
        y = layer.forward(y)

    if y is None:
        y: TensorGPU = layer.y
    # else: Nothing special.
    print(f"y | ({type(y)})")

    return y


def test_layers_gpu(model: PyDTNN_Model, dataset: np.ndarray) -> TensorGPU:

    print(f"test_layers_gpu - model")
    model.show()
    print(f"test_layers_gpu - model\n========")
    print(f"model.dtype: {model.dtype}")

    print(f"TYPES_DATA_CUDA[model.dtype]: {TYPES_DATA_CUDA[model.dtype]}")
    dtype = model.dtype
    model.cudnn_dtype = cudnn.cudnnDataType[TYPES_DATA_CUDA[model.dtype]]
    _dataset = TensorGPU(
        gpu_arr=gpuarray.empty(shape=dataset.shape, dtype=dtype),
        tensor_format=model.tensor_format, cudnn_dtype=model.cudnn_dtype)

    _dataset.ary.set(dataset)
    print(f"_dataset: {_dataset} | type(_dataset): {type(_dataset)} | _dataset.ary.shape: {_dataset.ary.shape}")

    model.y_batch = _dataset

    y: TensorGPU | None = forward_pydtnn_model(model, _dataset)

    return y


def test_layers(name: str, pytorch_model: TEST_PyTorch_Model, kwargs: Dict[str, Any], input_shape: Tuple[int, int, int],
                device: torch.device, dataset: np.ndarray, threshold: float, function_to_test_layers: Callable) -> None:

    print(pytorch_model)

    print("=======================\n== Converted version ==\n=======================")

    new_model: PyDTNN_Model = convert_model(model=pytorch_model, input_shape=input_shape,
                                            default_output_activation_layer=None,
                                            is_input_shape_in_format=True, **kwargs)

    new_model.mode = PyDTNN_Model.Mode.TRAIN
    # new_model.show()
    # new_model.dataset = dataset
    print("-----")
    # print("PyTorch model's forward method:")
    # graph = torch.fx.symbolic_trace(pytorch_model)
    # print(graph.code)
    # print("-----\n")

    # Must be only two layers: "Input" layer and the testing one.
    pydtnn_layer: LayerBase = new_model.layers[-1]
    print("=============================\n== Checking Dataset Values ==\n=============================")

    torch_dataset = torch.from_numpy(dataset).to(device)

    diff = torch_dataset.cpu().detach().numpy() - dataset
    print(f"Are equal: {are_all_zeros(diff)}")

    pytorch_state_dict = pytorch_model.layer.state_dict()

    pytorch_weights: None | torch.Tensor = pytorch_state_dict[PYTORCH_LAYER_WEIGHTS] if PYTORCH_LAYER_WEIGHTS in pytorch_state_dict else None
    pydtnn_weights: None | np.ndarray | TensorGPU = pydtnn_layer.weights

    if isinstance(pydtnn_weights, TensorGPU):
        pydtnn_weights: np.ndarray = pydtnn_weights.ary.get()

    there_are_pytorch_weigths = pytorch_weights is not None
    there_are_pydtnn_weights = pydtnn_weights is not None

    if there_are_pytorch_weigths and there_are_pydtnn_weights:
        pydtnn_weights = pydtnn_weights.T if name in TRANSPOSE_WEIGHTS_LAYERS else pydtnn_weights
        print(f"weigths are all zeros: {are_all_zeros(pytorch_weights.cpu().detach().numpy() - pydtnn_weights)}")

    pytorch_biases: None | torch.Tensor = pytorch_state_dict[PYTORCH_LAYER_BIASES] if PYTORCH_LAYER_BIASES in pytorch_state_dict else None
    pydtnn_biases: None | np.ndarray | TensorGPU = pydtnn_layer.biases

    if isinstance(pydtnn_weights, TensorGPU):
        pydtnn_biases: TensorGPU = pydtnn_biases.ary.get()

    there_are_pytorch_biases = pytorch_biases is not None
    there_are_pydtnn_biases = pydtnn_biases is not None

    if there_are_pytorch_biases and there_are_pydtnn_biases:
        print(f"biases are all zeros: {are_all_zeros(pytorch_biases.cpu().detach().numpy() - pydtnn_biases)}")

    print("=====================\n== Testing Forward ==\n=====================")
    print(f"pytorch_model: {pytorch_model}")

    pytorch_output: torch.Tensor = pytorch_model(torch_dataset)
    pydtnn_output: np.ndarray = function_to_test_layers(model=new_model, dataset=dataset)

    pytorch_output = pytorch_output.detach().to("cpu").numpy()

    diff = abs(pytorch_output) - abs(pydtnn_output)

    are_below_threshold = are_all_below_threshold(diff, threshold)
    print(f"Are equal: {are_all_zeros(diff)} || {name}")
    print(f"Are below the threshold ({threshold}): {are_below_threshold} || {name}")
    print(f"Min. value: {np.min(pydtnn_output)}")
    print(f"Max. value: {np.max(pydtnn_output)}")
    print(f"Mean of the values: {np.mean(pydtnn_output)}")
    print(f"Median of the values: {np.median(pydtnn_output)}")
    print(f"Max. difference between outputs: {np.max(diff)}")
    if not are_below_threshold:
        print(f"pytorch_output.shape: {pytorch_output.shape}")
        print(f"pydtnn_output.shape: {pydtnn_output.shape}")
        print(f"pytorch_output:\n{pytorch_output}\n[pytorch_output]")
        print(f"pydtnn_output:\n{pydtnn_output}\n[pydtnn_output]")
        print(f"pytorch_output - pydtnn_output:\n{diff}\n[pytorch_output - pydtnn_output]")

    print("=========================================\n")


def test_add_and_concat(name: str, pytorch_model: TEST_PyTorch_Model, kwargs: Dict[str, Any], input_shape: Tuple[int, int, int],
                        device: torch.device, dataset: np.ndarray, threshold: float = THRESHOLD) -> None:

    print(pytorch_model)

    print("=======================\n== Converted version ==\n=======================")

    # print("PyTorch model's forward method:")
    # graph = torch.fx.symbolic_trace(pytorch_model)
    # print(graph.code)

    print("-----\n")

    pydtnn_model: PyDTNN_Model = convert_model(model=pytorch_model, input_shape=input_shape,
                                               default_output_activation_layer=None,
                                               is_input_shape_in_format=True, **kwargs)

    # pydtnn_model.mode = ModelModeEnum.EVALUATE
    # pydtnn_model.show()
    torch_dataset = torch.from_numpy(dataset).to(device)
    print("-----")

    print("======================\n")

    print("=====================\n== Testing Forward ==\n=====================")
    print(f"pytorch_model: {pytorch_model}")

    pytorch_output, _ = pytorch_model(torch_dataset)
    pytorch_output: torch.Tensor
    pydtnn_model.dataset = dataset
    pydtnn_output = forward_pydtnn_model(pydtnn_model, dataset)
    pydtnn_output: np.ndarray

    pytorch_output = pytorch_output.detach().to("cpu").numpy()

    diff = abs(pytorch_output) - abs(pydtnn_output)
    # print(f"pytorch_output - pydtnn_output:\n{diff}")
    are_below_threshold = are_all_below_threshold(diff, threshold)
    print(f"Are equal: {are_all_zeros(diff)} || {name}")
    print(f"Are below the threshold ({threshold}): {are_below_threshold} || {name}")
    print(f"Min. value: {np.min(pydtnn_output)}")
    print(f"Max. value: {np.max(pydtnn_output)}")
    print(f"Median of the differences between outputs: {np.median(diff)}")
    print(f"Max. difference between outputs: {np.max(diff)}")
    if False and not are_below_threshold:
        print(f"pytorch_output:\n{pytorch_output}")
        print(f"pydtnn_output:\n{pydtnn_output}")
        print(f"pytorch_output - pydtnn_output:\n{diff}")
    print("=========================================\n")


def main():

    kwargs = KWARGS
    quarter_elements = prod((N, *SHAPE)) / 4

    device = torch.device("cpu") if kwargs["enable_gpu"] == False else torch.device("cuda")
    dataset_p = np.arange(quarter_elements, dtype=DTYPE) / 3
    dataset_p_int = np.arange(quarter_elements, dtype=DTYPE)
    dataset_n = np.arange(quarter_elements, dtype=DTYPE) * (-1 / 3)
    dataset_n_int = np.arange(quarter_elements, dtype=DTYPE) * (-1)

    dataset = np.concat([dataset_p, dataset_p_int, dataset_n, dataset_n_int]).reshape((N, *SHAPE))

    function_to_test_layers = (test_layers_gpu if device.type == "cuda" else forward_pydtnn_model)
    for name in DICT_SUPPORTED_LAYERS.keys():
        layer, threshold = DICT_SUPPORTED_LAYERS[name]
        model = TEST_PyTorch_Model(layer)
        print(f"Testing: {name}")
        print(f"{dataset.shape=}")
        print(f"{dataset.min()=}")
        print(f"{dataset.max()=}\n")

        test_layers(name=name, pytorch_model=model, kwargs=kwargs, input_shape=SHAPE,
                    device=device, dataset=np.copy(dataset), threshold=threshold,
                    function_to_test_layers=function_to_test_layers
                    )

    print("\n\n\n========================\n TESTING ADD AND CONCAT \n========================")

    for name, model in [("Addition", Addition_Test_PyTorch_Model()),
                        ("Concat", Concat_Test_PyTorch_Model()),
                        ]:
        print(f"Testing: {name}")
        test_add_and_concat(name=name, pytorch_model=model, kwargs=kwargs, input_shape=SHAPE, device=device, dataset=deepcopy(dataset))


if __name__ == "__main__":
    main()



from pydtnn.backends import PromoteToBackend
from pydtnn.utils import find_component


class Scheduler(PromoteToBackend):
    """
    Scheduler base class
    """

    def __init__(self, verbose: bool):
        self.verbose = verbose
        self.epoch_count = 0
        # NOTE: Only used in early_stopping and stop_at_loss.
        # NOTE (cont.): Since there are only 2 classes that uses this variable,
        #   I think it's not necessary to create an abstract class only to store this variable.
        self.stop_training: bool = False

    def __str__(self):
        return f"Scheduler {type(self).__name__}"

    def on_batch_begin(self, *args):
        pass

    def on_batch_end(self, *args):
        pass

    def on_epoch_begin(self, *args):
        pass

    def on_epoch_end(self, *args):
        pass

    def log(self, text: str):
        if self.verbose and self.model.comm_rank == 0:
            print(f"{self}: {text}")


def select(name: str) -> type[Scheduler]:
    assert __package__, "Package not found!"
    return find_component(__package__, name)



from pydtnn.backends.cpu.layers.layer import LayerCPU
from pydtnn.layers.abstract.pool_2d_layer import AbstractPool2DLayer
from pydtnn.utils.performance_models import im2col_time, col2im_time
from pydtnn.utils.tensor import TensorFormat
from pydtnn.utils.constants import ArrayShape

import numpy as np


class AbstractPool2DLayerCPU(AbstractPool2DLayer[np.ndarray], LayerCPU):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None):
        super().initialize(prev_shape, x)

        match self.model.tensor_format:
            case TensorFormat.NCHW:
                self.forward = self._forward_nchw_cython
                self.backward = self._backward_nchw_cython
            case TensorFormat.NHWC:
                self.forward = self._forward_nhwc_cython
                self.backward = self._backward_nhwc_cython
            case _:
                raise TypeError(f"Function: \'AbstractPool2DLayerCPU\'. Error:\n\tFormat: \'{self.model.tensor_format}\' not supported.")

        # I2C-based implementations have been temporarily discarded
        # setattr(self, "forward", self._forward_nchw_i2c)
        # setattr(self, "backward", self._backward_nchw_i2c)
        # setattr(self, "forward", self._forward_nhwc_i2c)
        # setattr(self, "backward", self._backward_nhwc_i2c)

        # The following variable is only for NCHW implementation (not for i2c implementation)
        y_shape = self.model.encode_shape((self.model.batch_size, self.co, self.ho, self.wo))
        # NOTE: This attribute only stores data, its value before the operation doesn't matter; it's initalized due avoid warnings in "LayerAndActivationBase.export".
        self.y = np.zeros(y_shape, dtype=self.model.dtype, order="C")

        self.fwd_time = \
            im2col_time(m=(self.kh * self.kw), n=(self.model.batch_size * self.ho * self.wo * self.ci),
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)
        self.bwd_time = \
            col2im_time(m=(self.kh * self.kw), n=(self.model.batch_size * self.ho * self.wo * self.ci),
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)

    def forward(self, x: np.ndarray) -> np.ndarray:
        msg = """This is a fake forward function. It will be masked on initialization by _forward_i2c or _forward_cg"""
        raise NotImplementedError(f"Class \'AbstractPool2DLayerCPU\'. Error: {msg}")

    def backward(self, dy: np.ndarray) -> np.ndarray:
        msg = """This is a fake backward function. It will be masked on initialization by _backward_i2c or _backward_cg"""
        raise NotImplementedError(f"Class \'AbstractPool2DLayerCPU\'. Error: {msg}")
    # ---

    def _forward_nchw_cython(self, x: np.ndarray) -> np.ndarray:
        raise NotImplementedError()

    def _backward_nchw_cython(self, dy: np.ndarray) -> np.ndarray:
        raise NotImplementedError()

    def _forward_nhwc_cython(self, x: np.ndarray) -> np.ndarray:
        raise NotImplementedError()

    def _backward_nhwc_cython(self, dy: np.ndarray) -> np.ndarray:
        raise NotImplementedError()



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def log_fwd_cython[T:_npDT](x: _npDT_1Dims[T], y: _npDT_1Dims[T]) -> None:
    """
    Args:
        x (npDT_1Dims): 1-dimensional input's array.
        y (npDT_1Dims): 1-dimensional array where the ouput is stored
    Returns:
        Nothing. The output is stored in "y".
    """


def log_bwd_cython[T:_npDT](dy: _npDT_1Dims[T], dx: _npDT_1Dims[T]) -> None:
    """
    Args:
        dy (npDT_1Dims): 1-dimensional input's array.
        dx (npDT_1Dims): 1-dimensional array where the output will be stored.
    Returns:
        Nothing. The output is stored in "dx".
    """



from pydtnn.layers.abstract.pool_2d_layer import AbstractPool2DLayer
from pydtnn.libs import libcudnn as cudnn
import pycuda.gpuarray as gpuarray   # type: ignore

from pydtnn.tracers.events import PYDTNN_OPS_EVENT, PYDTNN_OPS_EVENTS, PYDTNN_EVENT_FINISHED, PYDTNN_OPS_EVENT_enum
from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.performance_models import im2col_time, col2im_time
from pydtnn.layers.layer import ParameterException
from pydtnn.utils.constants import ArrayShape


class AbstractPool2DLayerGPU(AbstractPool2DLayer[TensorGPU], LayerGPU):
    """
    Provides common methods to Pool2DGPU classes.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # The following attributes will be initalized later.
        self.pool_desc = None # TODO: set CDNN descripor type
        self.ci: int = None  # type: ignore
        self.hi: int = None  # type: ignore
        self.wi: int = None  # type: ignore
        self.kh: int = None  # type: ignore
        self.kw: int = None  # type: ignore
        self.co: int = None  # type: ignore
        self.ci: int = None  # type: ignore
        self.ho: int = None  # type: ignore
        self.wo: int = None  # type: ignore

    def initialize_pool_2d_gpu(self, prev_shape: ArrayShape, x: TensorGPU, pool_mode: int) -> None:
        # pool_mode comes from cudnn.CudnnPoolingMode

        if not (self.vdilation == 1 and self.hdilation == 1):
            raise ParameterException(f"cuDNN does not support dilated pooling. vdilation: {self.vdilation}, hdilation: {self.hdilation}")

        nan_prop = cudnn.cudnnNanPropagation['CUDNN_NOT_PROPAGATE_NAN']

        self.pool_desc = cudnn.cudnnCreatePoolingDescriptor()
        cudnn.cudnnSetPooling2dDescriptor(self.pool_desc, pool_mode, nan_prop,
                                          self.kh, self.kw, self.vpadding, self.hpadding,
                                          self.vstride, self.hstride)
        # Get output dimensions
        _, _, self.ho, self.wo = cudnn.cudnnGetPooling2dForwardOutputDim(self.pool_desc, x.desc)
        self.shape = self.model.encode_shape((self.co, self.ho, self.wo))

        # Activations y
        y_gpu = gpuarray.empty((self.model.batch_size, *self.shape), self.model.dtype)
        self.y = TensorGPU(y_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        # Derivative dx
        dx_gpu = gpuarray.empty(self.x.ary.shape, self.model.dtype)
        self.dx = TensorGPU(dx_gpu, self.model.tensor_format, self.model.cudnn_dtype)

        self.fwd_time = \
            im2col_time(m=(self.kh * self.kw), n=(self.model.batch_size * self.ho * self.wo * self.ci),
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)
        self.bwd_time = \
            col2im_time(m=(self.kh * self.kw), n=(self.model.batch_size * self.ho * self.wo * self.ci),
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)

    def forward(self, x: TensorGPU) -> TensorGPU:
        alpha, beta = 1.0, 0.0
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.FORWARD_CUDNN)
        cudnn.cudnnPoolingForward(self.model.cudnn_handle, self.pool_desc, alpha,
                                  x.desc, x.ptr, beta,
                                  self.y.desc, self.y.ptr)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.y

    def backward(self, dy: TensorGPU) -> TensorGPU:
        alpha, beta = 1.0, 0.0
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, self.id * PYDTNN_OPS_EVENTS + PYDTNN_OPS_EVENT_enum.BACKWARD_CUDNN_DX)
        # Compute dx
        cudnn.cudnnPoolingBackward(self.model.cudnn_handle, self.pool_desc, alpha,
                                   self.y.desc, self.y.ptr,
                                   dy.desc, dy.ptr,
                                   self.x.desc, self.x.ptr,
                                   beta, self.dx.desc, self.dx.ptr)
        self.model.tracer.emit_event(PYDTNN_OPS_EVENT, PYDTNN_EVENT_FINISHED)
        return self.dx



# https://aclanthology.org/W18-2509.pdf


from collections.abc import Sequence

from pydtnn.layer_base import LayerBase
from pydtnn.layers.encoder_decoder import EncoderDecoder
from pydtnn.layers.input import Input


def iwslt(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=((1, 512, 512), (1, 512, 512))))
    _(EncoderDecoder(enc_layers=6, dec_layers=6, embedl=512, d_k=64, heads=8, d_ff=2048, dropout_rate=0.1))

    return model



import numpy as np

from pydtnn.backends.cpu.metrics.metric import MetricCPU
from pydtnn.metrics.categorical_mae import CategoricalMAE


class CategoricalMAECPU(CategoricalMAE[np.ndarray], MetricCPU):

    def compute(self, y_pred: np.ndarray, y_targ: np.ndarray) -> float:
        # return np.sum(np.absolute(1 - y_pred[np.arange(b), np.argmax(y_targ, axis=1)]))
        error = np.subtract(y_pred, y_targ, dtype=self.model.dtype)
        np.absolute(error, out=error, dtype=self.model.dtype)
        return np.mean(error, dtype=self.model.dtype)
